#include <sys/param.h>
#include <regex.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>


#include "extensions.h"
#include "userfile.h"

static struct aclmember *aclmembers = NULL;
static struct aclmember *acltail = NULL;

/* Have we read the ACL file yet? */
static int acl_read = 0;


/* PURPOSE   : Retrieve a named entry from the ACL                       */
/* ARGUMENTS : pointer to the keyword and a handle to the acl members    */
/* RETURNS   : pointer to the acl member containing the keyword or NULL  */
/*************************************************************************/

struct aclmember *
getaclentry(char *keyword, struct aclmember **next)
{
    do {
        if (!*next) {
                *next = aclmembers;
        } else
            *next = (*next)->next;
    } while ((*next && strcmp((*next)->keyword, keyword))
            );

    return (*next);
}

/*************************************************************************/
/* FUNCTION  : parseline (static)                                        */
/* PURPOSE   : Parse a single line into an ACL entry                     */
/* ARGUMENTS : Keyword for the entry                                     */
/*             strtok(NULL, "\t ") should return the first argument      */
/* RETURNS   : The struct aclmember * created (via malloc)               */
/*************************************************************************/

static struct aclmember *
parseline(char *keyword)
{
    struct aclmember *member;
    int cnt;
    char *ptr;
    char *findspc;

    member = (struct aclmember *) calloc(1, sizeof(struct aclmember));
    (void) strcpy(member->keyword, keyword);
    cnt = 0;
    while ((ptr = strtok(NULL, " \t")) != NULL) {
        if (cnt >= MAXARGS) {
            syslog(LOG_ERR,
                "Too many args (>%d) in config file: %s %s %s %s %s ...",
                MAXARGS - 1, member->keyword, member->arg[0],
                member->arg[1], member->arg[2], member->arg[3]);
            break;
        }
        member->arg[cnt] = ptr;
	while (findspc = strstr(member->arg[cnt], "[:space:]")) {
		*findspc = ' ';
		*++findspc = '\0';
		strcat(member->arg[cnt], findspc+8);
	}
	cnt++;
    }
    return member;
}

/*************************************************************************/
/* FUNCTION  : parseacl (static)                                         */
/* PURPOSE   : Parse the acl buffer into its components                  */
/* ARGUMENTS : A pointer to the acl buffer                               */
/* RETURNS   : nothing                                                   */
/*************************************************************************/

static int
parseacl(char *aclbuf)
{
    char *ptr,
     *aclptr = aclbuf,
     *line;
    struct aclmember *member;

    if (!aclbuf || !(*aclbuf))
        return (0);

    while (*aclptr != '\0') {
        line = aclptr;
        while (*aclptr && *aclptr != '\n')
            aclptr++;
        *aclptr++ = (char) NULL;

        /* deal with comments */
        if ((ptr = strchr(line, '#')) != NULL)
            /* allowed escaped '#' chars for path-filter (DiB) */
            if (*(ptr-1) != '\\')
                *ptr = '\0';
        ptr = strtok(line, " \t");
        if (ptr) {
            if (strcasecmp(ptr, "include") == 0) {
                ptr = strtok(NULL, " \t");
                if (ptr) {
                    if (!readacl(ptr))
                        syslog(LOG_ERR,
                            "Error including file %s in config file", ptr);
                } else {
                    syslog(LOG_ERR,
                        "Missing filename for `include' in config file");
                }
            } else {
                member = parseline(ptr);
                if (acltail)
                    acltail->next = member;
                acltail = member;
                if (!aclmembers)
                    aclmembers = member;
            }
        }
    }
    return (1);
}

/*************************************************************************/
/* FUNCTION  : readacl                                                   */
/* PURPOSE   : Read the acl into memory                                  */
/* ARGUMENTS : The pathname of the acl                                   */
/* RETURNS   : 0 if error, 1 if no error                                 */
/*************************************************************************/

int
readacl(char *aclpath)
{
    FILE *aclfile;
    struct stat finfo;
    char *aclbuf;

    if ((aclfile = fopen(aclpath, "r")) == NULL) {
        fprintf(stderr, "cannot open config file: %s\n", strerror(errno));
        exit (0);
    }
    if (fstat(fileno(aclfile), &finfo) != 0) {
        fprintf(stderr, "cannot fstat config file: %s\n", strerror(errno));
        (void) fclose(aclfile);
        exit (0);
    }
    if (finfo.st_size == 0) {
        aclbuf = (char *) calloc(1, 1);
    } else {
        if (!(aclbuf = (char *)malloc((unsigned) finfo.st_size + 1))) {
            fprintf(stderr, "could not malloc aclbuf (%d bytes)", finfo.st_size + 1);
            (void) fclose(aclfile);
            exit (0);
        }
        if (!fread(aclbuf, (size_t) finfo.st_size, 1, aclfile)) {
            fprintf(stderr, "error reading config file: %s", strerror(errno));
            aclbuf = NULL;
            (void) fclose(aclfile);
            exit (0);
        }
        *(aclbuf + finfo.st_size) = '\0';
    }
    (void) fclose(aclfile);
    return (parseacl(aclbuf));
}
/*-------------------------------------------------------------------------*/
